package com.javacodegeeks.spout;

import com.google.gson.Gson;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.reflect.TypeToken;
import org.apache.storm.spout.SpoutOutputCollector;
import org.apache.storm.task.TopologyContext;
import org.apache.storm.topology.OutputFieldsDeclarer;
import org.apache.storm.topology.base.BaseRichSpout;
import org.apache.storm.tuple.Fields;
import org.apache.storm.tuple.Values;
import org.apache.storm.utils.Utils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.Reader;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Map;

/**
 * A spout that reads flight data from a json file. Emits {@code "id", "index", "type", "source"}.
 */
public class FlightDataReaderSpout extends BaseRichSpout {
    private static final Logger LOG = LoggerFactory.getLogger(FlightDataReaderSpout.class);

    private SpoutOutputCollector collector;
    private Reader jsonReader;
    private boolean completed = false;
    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

    @Override
    public void close() {
        if (jsonReader != null)
            try {
                jsonReader.close();
            } catch (IOException e) {
                LOG.error(e.getLocalizedMessage());
            }
    }

    // Called when a task for this component is initialized within a worker on the cluster.
    @Override
    public void open(Map conf, TopologyContext context, SpoutOutputCollector collector) {
        final String flightDataFile = conf.get("flightData").toString();
        try {
            this.jsonReader = Files.newBufferedReader(Paths.get(flightDataFile));
        } catch (IOException e) {
            throw new RuntimeException("Error reading file [" + flightDataFile + "]");
        }
        this.collector = collector;
    }

    @Override
    public void nextTuple() {
        if (completed) {
            Utils.sleep(5000);
            return;
        }

        Gson gson = new Gson();
        Map<String, JsonElement> map = gson.fromJson(jsonReader,
                new TypeToken<Map<String, JsonElement>>() {
                }.getType());

        JsonArray acListArray = map.get("acList").getAsJsonArray();

        int count = 0;
        for (JsonElement flight : acListArray) {
            if (++count % 100 == 0) {
                LOG.info("Number of records emitted " + count);
            }

            JsonObject flightAsJsonObject = flight.getAsJsonObject();
            String id = flightAsJsonObject.get("Id").getAsString();

            LOG.debug("Processing flight record : " + flightAsJsonObject);

            flightAsJsonObject.remove("Id");
            processFSeen(flightAsJsonObject);
            addLocationAsGeoPoint(flightAsJsonObject);

            String flightAsJsonString = flightAsJsonObject.toString();
            LOG.debug("Json string of record to be emitted: " + flightAsJsonString);
            // The index under which data will be stored in ES
            Values values = new Values(id, "flight-storm", "_doc", flightAsJsonString);
            collector.emit(values);
        }
        completed = true;
    }

    private void processFSeen(JsonObject jsonObject) {
        String fSeenAsString = jsonObject.get("FSeen").getAsString().replaceAll("\\D+", "");
        Calendar fSeen = Calendar.getInstance();
        fSeen.setTimeInMillis(Long.parseLong(fSeenAsString));
        String fSeenTimeAsString = DATE_FORMAT.format(fSeen.getTime());
        LOG.debug("Adding timestamp " + fSeen + " as pretty string " + fSeenTimeAsString);
        jsonObject.addProperty("FSeen", fSeen.getTimeInMillis());
        jsonObject.addProperty("FSeenDateTime", fSeenTimeAsString);

//        long timestamp = Long.parseLong(fSeenAsString);
//        Instant instant = Instant.ofEpochMilli(timestamp) ;
//        LocalDateTime ldt = LocalDateTime.ofInstant(instant, ZoneOffset.UTC);
//        jsonObject.addProperty("FSeen", timestamp);
//        jsonObject.addProperty("FSeenDateTime", ldt.format(DateTimeFormatter.ISO_ZONED_DATE_TIME));
    }

    private void addLocationAsGeoPoint(JsonObject jsonObject) {
        if (jsonObject.get("Lat") != null && jsonObject.get("Long") != null) {
            JsonObject locationElement = new JsonObject();
            locationElement.addProperty("lat", jsonObject.get("Lat").getAsFloat());
            locationElement.addProperty("lon", jsonObject.get("Long").getAsFloat());
            jsonObject.add("location", locationElement);
        }
    }

    @Override
    public void declareOutputFields(OutputFieldsDeclarer declarer) {
        declarer.declare(new Fields("id", "index", "type", "source"));
    }
}